/*  filedata.cpp
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#include "filedata.h"

#include <stdexcept>
#include "functions.h"
#include "int.h"
#include "equation.h"
#include "files.h"
#include "ginacutils.h"
#include "kinematics.h"
#include "integralfamily.h"
#include "feynmanrules.h"

using namespace std;

namespace Reduze {

/// \todo denote filename in exceptions
/// \todo rethrow exceptions to include filename
/// \todo throw in outfile in case of failures
/// \todo maybe check for failure other than eof in infile get members

OutFileData::OutFileData(const std::string& filename,
		const std::string& format_str) :
	filename_(filename), tmp_filename_(filename + ".tmp"), empty_(true) {
	if (format_str == "mma")
		format_ = MMA;
	else if (format_str == "maple")
		format_ = MAPLE;
	else if (format_str == "form")
		format_ = FORM;
	else if (format_str == "reduze")
		format_ = Reduze;
	else
		throw std::runtime_error("Unknown output format " + format_str);
	LOGXX("opening " << tmp_filename_);
	raw_ostream_.open(tmp_filename_.c_str());
	if (!raw_ostream_.good())
		throw std::runtime_error("can't open file " + tmp_filename_);

	if (format_ == MMA) {
		raw_ostream_ << "{\n";
	} else if (format_ == MAPLE) {
		string listname = get_filename_without_directory(filename);
		listname = get_filename_without_suffix(listname);
		listname = to_safe_variable_name(listname);
		raw_ostream_ << listname << " := [\n\n";
	}
}

OutFileData::~OutFileData() {
	close();
}

void OutFileData::close() {
	if (raw_ostream_.is_open()) {
		if (!empty_)
			raw_ostream_ << "\n";
		if (format_ == MMA) {
			raw_ostream_ << "}\n";
		} else if (format_ == MAPLE) {
			raw_ostream_ << "]\n";
		}
		raw_ostream_.close();
		rename(tmp_filename_, filename_);
	}
}

void OutFileData::write_comment(const std::string& s) {
	/// \todo: implement me
}

std::string OutFileData::default_filename_suffix(const std::string& format) {
	if (format == "mma")
		return ".m";
	else if (format == "maple")
		return ".mpl";
	else if (format == "form")
		return ".inc";
	else if (format == "reduze")
		return ".reduze";
	else
		throw std::runtime_error("unknown output format " + format);
}

size_t InFileData::file_size() {
	size_t pos = raw_istream_.tellg();
	raw_istream_.seekg(0);
	size_t begin = raw_istream_.tellg();
	ASSERT(begin == 0);
	raw_istream_.seekg(0, ios::end);
	size_t end = raw_istream_.tellg();
	raw_istream_.seekg(pos);
	return (end - begin);
}

InFileData::InFileData(const std::string& filename) {
	raw_istream_.open(filename.c_str());
	if (!raw_istream_.good())
		throw runtime_error("can't open file " + filename);
}

InFileData& InFileData::content_line(std::string& l) {
	l.clear();
	while (l.empty() && getline(raw_istream_, l)) {
		size_t to = std::min(l.find("#"), l.find("//"));
		l = chop_whitespaces(l.substr(0, to));
	}
	return *this;
}

// OutFileINTs

OutFileINTs& OutFileINTs::operator<<(const INT& i) {
	this->OutFileData::operator<<(i);
	return *this;
}

OutFileINTs& OutFileINTs::operator<<(const std::set<INT>& is) {
	for (set<INT>::const_iterator i = is.begin(); i != is.end(); ++i)
		this->OutFileData::operator<<(*i);
	return *this;
}

// InFileINTs
namespace {
// removes simple Mathematica (* comment *), brackets and lead/trail whitespaces
string remove_mma_decoration(const string& sin) {
	string s(sin);
	while (true) {
		size_t to = s.find("*)");
		size_t from = s.find("(*");
		to = (to == string::npos ? 0 : to);
		from = (from == string::npos || from > to ? 0 : from);
		if (from == 0 && to == 0)
			break;
		else
			s.replace(from, to + 2 - from, to + 2 - from, ' ');
	}
	return chop_whitespaces(replace_chars(s, "{}", ' '));
}
}

InFileINTs& InFileINTs::get(INT& i) {
	string l;
	do {
		content_line(l);
		l = remove_mma_decoration(replace_chars(l, "{}", ' '));
	} while (l.empty() && good());
	if (!l.empty()) {
		try {
			i = INT(l);
		} catch (exception&) {
			i = INT::from_mma(l);
		}
	}
	return *this;
}

void InFileINTs::get_all(set<INT>& ints) {
	IntegralFamily dummy;
	INT i(&dummy);
	while (get(i))
		ints.insert(i);
}

// OutFileLinearCombinations

OutFileLinearCombinations& OutFileLinearCombinations::operator<<(
		const LinearCombination& lc) {
	this->OutFileData::operator<<(lc);
	return *this;
}

// InFileLinearCombinations

bool InFileLinearCombinations::check_filetype(const string& filename) {
	InFileLinearCombinations in(filename);
	try {
		LinCombHLight lc;
		in.get(lc); // throw if not a lin comb, silent failure for empty file
	} catch (exception&) {
		return false;
	}
	return true;
}

template<class TSet>
void InFileLinearCombinations::find_INT(const std::string& filename, TSet& ints) {
	InFileLinearCombinations in(filename);
	LinCombHLight lc;
	bool discard_coeff = true;
	while (in.get(lc, discard_coeff))
		lc.find_INT(ints);
}

InFileLinearCombinations::InFileLinearCombinations(const std::string& filename) :
	InFileData(filename) {
	Files* files = Files::instance();
	symbols_ = files->kinematics()->kinematic_invariants_and_dimension();
	if (is_readable_file(files->get_filename_feynmanrules())) {
		GiNaC::lst s = files->feynmanrulesglobals()->coupling_constants();
		symbols_ = add_lst(symbols_, s);
		symbols_.append(files->feynmanrulesglobals()->Nc());
	}
}

InFileLinearCombinations& InFileLinearCombinations::get(LinearCombination& lc) {
	string lcstr, l;
	do {
		content_line(l);
		lcstr += (l.empty() ? "" : l + '\n');
	} while (l != ";" && good());
	if (!l.empty()) {
		ICharArrayBuffer ib(lcstr.c_str(), lcstr.size());
		istream is(&ib);
		lc = LinearCombination(is, symbols_);
	}
	return *this;
}

InFileLinearCombinations& InFileLinearCombinations::get(LinCombHLight& lc,
		bool discard_coeff) {
	string lcstr, l;
	do {
		content_line(l);
		lcstr += (l.empty() ? "" : l + '\n');
	} while (l != ";" && good());
	if (!l.empty()) {
		ICharArrayBuffer ib(lcstr.c_str(), lcstr.size());
		istream is(&ib);
		lc = LinCombHLight(is, discard_coeff);
	}
	return *this;
}

// OutFileEquations

OutFileEquations& OutFileEquations::operator<<(const Identity& eq) {
	this->OutFileData::operator<<(eq);
	return *this;
}

OutFileEquations& OutFileEquations::operator<<(const EquationHLight& eq) {
	if (format() == Reduze)
		write_reduze(eq);
	else
		this->OutFileData::operator<<(Identity(eq));
	return *this;
}

// InFileEquations

bool InFileEquations::check_filetype(const string& filename) {
	InFileEquations in(filename);
	try {
		Identity id;
		in.get(id); // throw if not an equation, silent failure for empty file
	} catch (exception&) {
		return false;
	}
	return true;
}

InFileEquations::InFileEquations(const std::string& filename) :
	InFileData(filename) {
	Files* files = Files::instance();
	symbols_ = files->kinematics()->kinematic_invariants_and_dimension();
	if (is_readable_file(files->get_filename_feynmanrules())) {
		GiNaC::lst s = files->feynmanrulesglobals()->coupling_constants();
		symbols_ = add_lst(symbols_, s);
		symbols_.append(files->feynmanrulesglobals()->Nc());
	}
}

InFileEquations& InFileEquations::get(Identity& eq) {
	string lcstr, l;
	do {
		content_line(l);
		lcstr += (l.empty() ? "" : l + '\n');
	} while (l != ";" && good());
	if (!l.empty()) {
		ICharArrayBuffer ib(lcstr.c_str(), lcstr.size());
		istream is(&ib);
		LinearCombination lc(is, symbols_);
		if (!lc.name().empty())
			throw runtime_error("found name but expected unnamed equation");
		lc.swap_terms(eq);
	}
	return *this;
}

InFileEquations& InFileEquations::get(EquationHLight& eq) {
	string lcstr, l;
	do {
		content_line(l);
		lcstr += (l.empty() ? "" : l + '\n');
	} while (l != ";" && good());
	if (!l.empty()) {
		ICharArrayBuffer ib(lcstr.c_str(), lcstr.size());
		istream is(&ib);
		LinCombHLight lc(is);
		if (!lc.name().empty())
			throw runtime_error(
					"found name but expected unnamed equation in file");
		lc.swap_terms(eq);
	}
	return *this;
}

}

// explicit instantiation
template void
Reduze::InFileLinearCombinations::find_INT<std::set<Reduze::INT> >(
		const std::string& filename, std::set<Reduze::INT>& ints);
template void
Reduze::InFileLinearCombinations::find_INT<std::multiset<Reduze::INT> >(
		const std::string& filename, std::multiset<Reduze::INT>& ints);
#ifdef USE_HASH_TABLE
template void
Reduze::InFileLinearCombinations::find_INT<tr1::unordered_set<Reduze::INT> >(
		const std::string& filename, tr1::unordered_set<Reduze::INT>& ints);
#endif /* USE_HASH_TABLE */

