/*  sectormappings.cpp
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#include "sectormappings.h"

#include "kinematics.h"
#include "int.h"
#include "equationlist.h"
#include "identitygenerator.h"
#include "functions.h"
#include "files.h"
#include "ginacutils.h"
#include "yamlutils.h"
#include "graph.h"

using namespace std;
using namespace GiNaC;

namespace Reduze {

bool SectorMappings::no_explicit_shift_verification_ = false;

// constructors

SectorMappings::SectorMappings(const IntegralFamily* ic) :
	integralfamily_(ic) {
}

SectorMappings::SectorMappings(const CrossedIntegralFamily* ic) :
	integralfamily_(ic) {
	string sn = ic->source_integralfamily()->name();
	/*	const SectorMappings* sm = Files::instance()->sectormappings(sn);
	 zero_sectors_ = sm->zero_sectors_;
	 sectors_without_graph_ = sm->sectors_without_graph_;
	 */
}

void SectorMappings::setup(std::set<SectorGL>& shift_targets,
		std::set<SectorGL>& local_shift_targets,
		const SetupSectorMappingsOptions& options, std::map<Sector, std::list<
				std::map<int, int> > >& cached_node_perm) {

	bool find_sector_relations = options.find_sector_relations_;
	bool find_sector_symmetries = options.find_sector_symmetries_;
	bool find_zero_sectors = options.find_zero_sectors_;
	bool find_unphys_zero_sectors = options.find_graphless_zero_sectors_;
	bool minimize_by_twists = options.minimize_graphs_by_twists_;
	bool construct_minimal_graphs = options.construct_minimal_graphs_;
	bool minimize_target_crossings = options.minimize_target_crossings_;
	bool allow_general_crossings = options.allow_general_crossings_;
	bool find_crossed_relations = options.allow_general_crossings_;

	local_shift_targets.clear();

	const Files* files = Files::instance();
	const IntegralFamily* ic = integralfamily_;
	const int max_sec_id = ic->max_sector_id();
	const bool allow_long_zero_test =
			ic->kinematics()->external_momenta().nops() != 0;
	LOG("Setup sector mappings for " << ic->name());

	const list<Crossing>& oc = ic->kinematics()->ordered_crossings();
	LOGX("Ordered crossings:");
	for (list<Crossing>::const_iterator c = oc.begin(); c != oc.end(); ++c) {
		const map<Crossing, Crossing>& xmap =
				ic->kinematics()->equivalent_crossings();
		VERIFY(xmap.find(*c) != xmap.end());
		YAML::Emitter ye;
		ye << *c;
		LOGX ("  " << ye.c_str() << "\n"
				<< "is equivalent to '" << xmap.find(*c)->second.name() << "'");
	}
	LOGX("");

	YAML::Emitter e;
	ic->print(e);
	LOGX(e.c_str());
	LOGX("");
	LOGX("Rules scalar products to invariants:\n" << ic->kinematics()->rules_sp_to_invariants());
	LOGX("Rules invariants to scalar products:\n" << ic->kinematics()->find_rules_invariants_to_sp());
	LOGX("Rules scalar products to propagators:\n" << ic->rules_sp_to_prop());
	LOGX("");

	LOG("Maximal sector id is " << max_sec_id << ".");
	LOG(max_sec_id + 1 << " sectors occur in total");

	// sector tree (using global permutation symmetry)

	LOG("Finding non-equivalent sectors using permutation symmetries:");
	bool perm_sym = !ic->permutation_symmetries().empty();
	set<Sector> sectors;
	set<int> sector_ids;
	for (int i = 0; i <= max_sec_id; ++i) {
		int id = perm_sym ? ic->get_equivalent(i) : i;
		sectors.insert(Sector(ic, id));
		sector_ids.insert(sector_ids.end(), id);
	}
	// print non-equivalent sectors to log-file
	LOGX("\nNon-equivalent sectors:");
	LOGX(ic->t_tree_nice_string(sector_ids));
	LOG(sector_ids.size() << " non-equivalent sectors were found");

	// graphs of sectors
	// sectors with vanishing jacobian determinant of the loop momenta matrix are added to zero_sectors_

	map<Sector, Graph> graph_by_sector;
	find_graphs(graph_by_sector, sectors, minimize_by_twists,
			construct_minimal_graphs, true);
	//find_graphs_from_top(graph_by_sector, sectors, minimize_by_twists);
	LOGX("\nNo graph sectors:");
	LOGX(ic->t_tree_nice_string(sectors_without_graph()));

	// canonical labeling

	LOG("Finding canonical labels:");
	ProgressBar pbar(2, "canonical labels:", graph_by_sector.size());
	pbar.start();
	set<SectorGL> local_sectors;
	for (map<Sector, Graph>::const_iterator s = graph_by_sector.begin(); s
			!= graph_by_sector.end(); ++s) {
		local_sectors.insert(SectorGL(s->first, s->second,
				s->second.find_canonical_graph_label()));
		pbar.print();
	}
	pbar.end();

	// list of sectors by their shift target (all sectors are from this family, target not included in the set)
	map<SectorGL, list<SectorGL> > local_targets_by_equiv;

	// try to shift the 'local_sectors' to the shift_targets or to them self
	// updates the 'shift_targets'
	// sets the 'local_shift_targets'
	if (find_sector_relations) {
		find_sector_shifts(shift_targets, local_sectors,
				local_targets_by_equiv, minimize_target_crossings,
				find_zero_sectors, true, allow_general_crossings,
				cached_node_perm, true);
	} else {
		// move local sectors to local_shift_targets
		map<SectorGL, list<SectorGL> >::iterator in =
				local_targets_by_equiv.end();
		list<SectorGL> empty;
		for (set<SectorGL>::iterator s = local_sectors.begin(); s
				!= local_sectors.end();) {
			in = local_targets_by_equiv.insert(in, make_pair(*s, empty));
			local_sectors.erase(s++);
		}
	}

	// zero test for the local_targets
	if (find_zero_sectors) {
		LOGN("Finding zero sectors of sectors with a graph:");
		map<Sector, list<Sector> > zeros;
		map<Sector, list<Sector> >::iterator in = zeros.end();
		map<SectorGL, list<SectorGL> >::const_iterator s;
		list<SectorGL>::const_iterator t;
		for (s = local_targets_by_equiv.begin(); s
				!= local_targets_by_equiv.end(); ++s) {
			in = zeros.insert(in, make_pair(s->first.sector_, list<Sector> ()));
			for (t = s->second.begin(); t != s->second.end(); ++t)
				in->second.push_back(t->sector_);
		}
		this->find_zero_sectors(zeros, allow_long_zero_test);
		LOG("Found " << zero_sectors_.size() << " zero sectors (isomorphics included).");
	}

	if (find_zero_sectors && find_unphys_zero_sectors) {
		LOGN("Finding zero sectors for sectors without a graph:");
		map<Sector, list<Sector> > zeros;
		map<Sector, list<Sector> >::iterator in = zeros.end();
		set<int>::const_iterator s = sectors_without_graph_.begin();
		int num_phys_zeros = zeros.size();
		for (; s != sectors_without_graph_.end(); ++s)
			in = zeros.insert(in, make_pair(Sector(ic, *s), list<Sector> ()));
		this->find_zero_sectors(zeros, allow_long_zero_test);
		LOG("Found additional " << zeros.size() - num_phys_zeros << " zero sectors.");
	}

	set<int> non_zero_shift_targets;
	set<GraphLabel> local_target_labels;
	for (map<SectorGL, list<SectorGL> >::const_iterator s =
			local_targets_by_equiv.begin(); s != local_targets_by_equiv.end(); ++s) {
		if (zero_sectors_.find(s->first.sector_.id()) == zero_sectors_.end()) {
			non_zero_shift_targets.insert(s->first.sector_.id());
			local_target_labels.insert(s->first.labels_.first);
		}
		local_shift_targets.insert(s->first);
	}
	LOGX("\nNon-vanishing shift targets:");
	LOGX(ic->t_tree_nice_string(non_zero_shift_targets));
	LOG("Found " << non_zero_shift_targets.size() << " non-vanishing shift target sectors");
	// the non-vanishing local shift targets can still contain graphs which are isomorphic
	// but a shift between them would have |det| != 1. Log info:
	int num_same_graph = non_zero_shift_targets.size()
			- local_target_labels.size();
	LOG("  " << local_target_labels.size() << " of them "
			<< (num_same_graph == 1 ? "has" : "have")
			<< " different graphs.");

	// symmetry shifts

	if (find_sector_symmetries) {
		LOG("Finding symmetry shift relations:");
		SectorMappings::find_symmetry_shifts(local_shift_targets,
				cached_node_perm);
	}
	//LOG("");

	// pure crossing relations: crossed to uncrossed sector
	// setup of uncrossed mappings done, it is safe now to construct
	// crossed mappings based on them (implicitely via Files::sectormappings())
	// and enrich them

	if (find_crossed_relations) {
		LOG("Finding crossed sector relations for non-zero shift targets");
		// find crossed counter parts for sector mappings (one per equiv. class)
		list<SectorMappings*> ms;
		//const list<Crossing>& xs = ic->kinematics()->ordered_crossings();
		//for (list<Crossing>::const_iterator x = xs.begin(); x != xs.end(); ++x)
		// it is enough to consider minimal crossing representatives as sources
		set<Crossing> xs = ic->kinematics()->equivalent_crossing_classes();
		for (set<Crossing>::const_iterator x = xs.begin(); x != xs.end(); ++x)
			if (!x->is_equivalent_to_identity()) {
				string xfname = x->name_for_crossed_family(integralfamily_);
				ms.push_back(Files::instance()->sectormappings(xfname));
			}
		// shift crossed versions of target sectors to uncrossed ones
		ProgressBar pbar(2, "crossed relations:", local_shift_targets.size());
		pbar.start();
		for (set<SectorGL>::const_iterator s = local_shift_targets.begin(); s
				!= local_shift_targets.end(); ++s) {
			for (list<SectorMappings*>::iterator m = ms.begin(); m != ms.end(); ++m) {
				// construct crossed sector and its graph
				Sector xsec((*m)->integralfamily(), s->sector_.id());
				map<Sector, Graph> gs;
				set<Sector> xsecset;
				xsecset.insert(xsec);
				(*m)->get_graphs(gs, xsecset, minimize_by_twists,
						construct_minimal_graphs, false);
				VERIFY(gs.find(xsec) != gs.end());
				Graph g = gs.find(xsec)->second;
				// find shift
				set<SectorGL> from, to;
				from.insert(SectorGL(xsec, g, g.find_canonical_graph_label()));
				to.insert(*s);
				map<SectorGL, list<SectorGL> > dummy;
				(*m)->find_sector_shifts(to, from, dummy, true, false, true,
						true, cached_node_perm, false);
				map<int, Sector>::const_iterator rel = //
						(*m)->sector_relations().find(s->sector_.id());
				if (rel != (*m)->sector_relations().end()) {
					const Sector& target = rel->second;
					if (Crossing::to_minimal_crossing(target) == //
							Crossing::to_minimal_crossing(xsec)) {
						// delete shift, is covered by crossing equivalences
						(*m)->remove_shift(s->sector_.id());
					} else {
						LOGX(xsec << "=>" << s->sector_ << ": " << target);
					}
				}
			}
			pbar.print();
		}
		pbar.end();
	}

	// print graphs as dot files in standard directory ./graphs/family_name/

	const string graph_dir = files->get_graph_directory(ic);
	LOG("Printing graphs from sectors in " << graph_dir << ":");
	remove_directory_with_files(graph_dir);
	make_directory(graph_dir);
	const string all_dir = graph_dir + "all/";
	const string shift_targets_dir = graph_dir + "shift_targets/";
	make_directory(all_dir);
	make_directory(shift_targets_dir);

	string dot_file_prefix = all_dir + safe_filename(ic->name());
	string dot_file_prefix2 = shift_targets_dir + safe_filename(ic->name());
	string ps_file = safe_filename(ic->name()) + ".ps";
	string ps_file2 = safe_filename(ic->name()) + "_shift_targets.ps";
	for (map<Sector, Graph>::const_iterator s = graph_by_sector.begin(); s
			!= graph_by_sector.end(); ++s)
		s->second.print_dot(all_dir + s->first.get_safe_string_for_filename(
				true) + ".dot");
	for (set<SectorGL>::const_iterator s = local_shift_targets.begin(); s
			!= local_shift_targets.end(); ++s)
		//if (!is_zero_sector_non_static(s->sector_.id()))
		s->graph_.print_dot(shift_targets_dir
				+ s->sector_.get_safe_string_for_filename(true) + ".dot");

	LOG("To generate a postscript file from the graphs in the dot format type e.g.");
	LOG("  dot -Tps -o " << ps_file << " " << dot_file_prefix << "_*.dot");
	LOG("  dot -Tps -o " << ps_file2 << " " << dot_file_prefix2 << "_*.dot");
	//LOG("");
	LOG("Done setup sector mappings for " << ic->name());
}

bool SectorMappings::is_zero(const Sector& s) {
	return get_shifted_sector(s).is_obviously_zero();
}

// get equivalent and shifted sectors

INT SectorMappings::get_equivalent_or_unique_zero(const INT& in) {
	// minimal crossing, permutation symmetries
	INT i = Crossing::to_minimal_crossing(in);
	// exploit permutation symmetries and zero sectors of uncrossed family
	pair<INT, Crossing> unx = Crossing::uncross(i);
	// permutation symmetry
	INT inox = unx.first.get_equivalent();
	// map zero integrals to zero sector integral
	INT z = INT(inox.integralfamily(), 0);
	if (z.is_obviously_zero()) { // #loops>0, ID=0 really is zero
		if (inox.is_obviously_zero())
			return z;
		const SectorMappings* m = Files::instance()->sectormappings(
				inox.integralfamily()->name());
		if (m->zero_sectors_.find(inox.id()) != m->zero_sectors_.end())
			return z;
	}
	// all done
	return unx.second.transform(inox);
}

Sector SectorMappings::get_shifted_sector(const Sector& secin, bool recurse) {
	// minimal crossing, permutation symmetry, zero sectors
	INT i = get_equivalent_or_unique_zero(INT(secin));
	// check for explicit shifts of crossed sector (to uncrossed version of same sector)
	if (i.integralfamily()->is_crossed()) {
		const SectorMappings* xm = Files::instance()->sectormappings(
				i.integralfamily()->name());
		map<int, Sector>::const_iterator to =
				xm->sector_relations_.find(i.id());
		if (to != xm->sector_relations_.end()) {
			if (!recurse)
				return get_equivalent_or_unique_zero(INT(to->second)).get_sector();
			return get_shifted_sector(to->second, recurse);
		}
	}
	// use shifts from uncrossed family
	pair<INT, Crossing> unx = Crossing::uncross(i);
	const SectorMappings* m = Files::instance()->sectormappings(
			unx.first.integralfamily()->name());
	int id = unx.first.id();
	map<int, Sector>::const_iterator targetnox = m->sector_relations_.find(id);
	if (targetnox != m->sector_relations_.end()) {
		Sector target = unx.second.transform(targetnox->second);
		if (!recurse)
			return get_equivalent_or_unique_zero(INT(target)).get_sector();
		return get_shifted_sector(target, recurse);
	}
	// nothing to reduce
	return i.get_sector();
}

bool SectorMappings::search_corner_reduction_to_zero(int sector_id,
		bool allow_longtest) const {
	Sector sector(integralfamily_, sector_id);
	LOGXX("Checking if sector " << sector << " is zero");
	// for tree level integral family, t=0 is not zero
	if (sector.is_obviously_zero())
		return LOGXX("Is trivial zero sector"), true;
	if (sector.is_obviously_non_zero())
		return LOGXX("Is trivial tree level (non-zero) sector"), false;
	if (zero_sectors_.find(sector_id) != zero_sectors_.end())
		return LOGXX("Is already known to be zero"), true;
	INT corner(integralfamily_, sector_id);
	INT equiv = corner.get_equivalent();
	if (equiv != corner) {
		WARNING("Performing zero test for equivalent integral " << equiv
				<< " instead of " << corner);
		corner = equiv;
		sector = Sector(integralfamily_, equiv.id());
	}
	int t = corner.t();
	set<INT> seeds;
	seeds.insert(corner);
	IdentityList idlist;
	IdentityGeneratorOptions options;
	options.set_eliminate_subsectors_by_shift(false); // sectormappings aren't known yet
	static IdentityGenerator generator(options); // static gives great speedup here
	generator.generate(seeds, idlist, "ibp");
	idlist.remove_empty_equations();
	if (idlist.size() == 0)
		WARNING("Empty identity list in zero test for sector " << sector_id);
	LOGXX(idlist);
	idlist.reduze();
	if (idlist.gives_reduction_zero(corner))
		return LOGXX("Is zero since corner reduces to zero"), true;
	// increasing s2 can't help for maximal sector
	if (sector_id == integralfamily_->max_sector_id())
		return LOGXX("Treat maximal sector as non-zero"), false;
	// test with more equations, s2 -> s2 + 1
	if (allow_longtest) {
		SeedGeneratorOptions seed_options;
		seed_options.set_reduce_to_equivalent(true);
		SeedGenerator seed_generator(integralfamily_, seed_options);
		set<INT> seeds2;
		seed_generator.generate(sector_id, t, t, 1, 1, seeds2);
		IdentityList idlist2;
		generator.generate(seeds2, idlist2, "ibp");
		idlist2.splice(idlist);
		idlist2.remove_empty_equations();
		idlist2.reduze();
		seeds2.insert(seeds2.begin(), corner);
		if (idlist2.gives_reduction_zero(
				set<INT> (seeds2.begin(), seeds2.end()))) {
			return LOGXX("Is zero since extended seeds reduce to zero"), true;
		}
	}
	return LOGXX("Not found to reduce to zero"), false;
}

void SectorMappings::get_graphs(std::map<Sector, Graph>& graph_by_sector,
		const std::set<Sector>& sectors, bool minimize_by_twists,
		bool construct_minimal_graphs, bool verbose) const {
	string twistonoff = (minimize_by_twists ? "on" : "off");
	string minonoff = (construct_minimal_graphs ? "on" : "off");
	ProgressBar pbar(2, "finding graphs: (twists " + twistonoff
			+ " / construct minimal " + minonoff + ")", sectors.size());
	if (verbose)
		pbar.start();
	pair<map<Sector, Graph>::iterator, bool> inserted;
	for (set<Sector>::const_iterator s = sectors.begin(); s != sectors.end(); ++s) {
		if (verbose)
			pbar.print();
		bool det_non_zero;
		Graph g;
		if (s->search_graph(g, det_non_zero, construct_minimal_graphs)) {
			if (minimize_by_twists)
				g.minimize_by_twists();
			inserted = graph_by_sector.insert(make_pair(*s, Graph()));
			ASSERT(inserted.second);
			inserted.first->second.swap(g);
		}
	}
	if (verbose)
		pbar.end();
}

void SectorMappings::find_graphs(std::map<Sector, Graph>& graph_by_sector,
		const std::set<Sector>& sectors, bool minimize_by_twists,
		bool construct_minimal_graphs, bool verbose) {
	if (verbose)
		LOG("Finding graphs of sectors (twists " << (minimize_by_twists ? "on" : "off")
				<< " / construct minimal " << (construct_minimal_graphs ? "on" : "off")
				<< "):");
	ProgressBar pbar(2, "finding graphs", sectors.size());
	if (verbose)
		pbar.start();
	pair<map<Sector, Graph>::iterator, bool> inserted;
	int t = -1, col = 1, num_zeros = 0;
	for (set<Sector>::const_iterator s = sectors.begin(); s != sectors.end(); ++s) {
		if (verbose)
			pbar.print();
		int id = s->id();
		if (s->t() != t) {
			t = s->t();
			col = 1;
			//LOGN(endl << "   t = " << t << ":\t\b\b\b\b" << flush);
		} // else if ((col++) % 60 == 0) {
		//	LOGN (endl << "\t\t\b\b\b\b" << flush);
		//}
		set<int> sub = integralfamily_->get_subsector_equivalents(id, false);
		set<int>::iterator n = sub.begin();
		for (; n != sub.end(); ++n)
			if (sectors_without_graph_.find(*n) != sectors_without_graph_.end())
				break;
		bool det_non_zero = true;
		Graph g;
		if (n != sub.end() || !(s->search_graph(g, det_non_zero,
				construct_minimal_graphs))) {
			if (det_non_zero) {
				sectors_without_graph_.insert(id);
				//LOGN("n" << flush);
			} else {
				++num_zeros;
				zero_sectors_.insert(id); // det = 0
				//LOGN("0" << flush);
			}
		} else {
			if (minimize_by_twists)
				g.minimize_by_twists();
			inserted = graph_by_sector.insert(make_pair(*s, Graph()));
			ASSERT(inserted.second);
			inserted.first->second.swap(g);
			//LOGN("g" << flush);
		}
	}
	if (verbose) {
		pbar.end();
		LOG("Found " << graph_by_sector.size() << " graphs");
	}
}

// Baustelle
//void SectorMappings::find_graphs_from_top(
//		std::map<Sector, Graph>& graph_by_sector, const std::set<Sector>& secs,
//		bool minimize_by_twists) {
//
//	LOG("Finding graphs of sectors (twists " << (minimize_by_twists ? "on" : "off") << "):");
//	set<Sector, greater<Sector> > sectors(secs.begin(), secs.end()); // copy
//	ProgressBar pbar(2, "finding graphs", sectors.size());
//	pbar.start();
//	pair<map<Sector, Graph>::iterator, bool> inserted;
//	int num_zeros = 0;
//	for (set<Sector, greater<Sector> >::iterator s = sectors.begin(); s
//			!= sectors.end(); ++s) {
//		VERIFY(*integralfamily_ == *s->integralfamily());
//		cerr << "find graph for sector " << *s << endl;
//		pbar.print();
//		int id = s->id();
//		bool det_non_zero = true;
//		Graph g;
//		if (!(s->search_graph(g, det_non_zero))) {
//			if (det_non_zero) {
//				sectors_without_graph_.insert(id);
//			} else {
//				++num_zeros;
//				zero_sectors_.insert(id); // det = 0
//			}
//		} else {
//			if (minimize_by_twists)
//				g.minimize_by_twists();
//			inserted = graph_by_sector.insert(make_pair(*s, Graph()));
//			ASSERT(inserted.second);
//			inserted.first->second.swap(g);
//			set<int> subs = integralfamily_->get_subsector_equivalents(
//					s->id(), true);
//			for (set<Sector, greater<Sector> >::iterator ss = s; ss
//					!= sectors.end();)
//				if (subs.find(ss->id()) != subs.end())
//					sectors.erase(ss++);
//				else
//					++ss;
//		}
//	}
//	pbar.end();
//	LOG("Found " << graph_by_sector.size() << " graphs");
//}

void SectorMappings::find_canonical_graph_labels(std::map<Sector, std::pair<
		GraphLabel, GraphRelabeling> >& labels, //
		const std::map<Sector, Graph>& sectors) {
	LOG("  Find canonical labels:");
	ProgressBar pbar(2, "canonical labels:", sectors.size());
	pbar.start();
	map<Sector, Graph>::const_iterator s;
	map<Sector, pair<GraphLabel, GraphRelabeling> >::iterator hint =
			labels.begin();
	for (s = sectors.begin(); s != sectors.end(); ++s) {
		hint = labels.insert(hint, make_pair(s->first,
				s->second.find_canonical_graph_label()));
		pbar.print();
	}
	pbar.end();
}

void SectorMappings::find_sector_shifts(
		std::set<SectorGL>& minimal, //
		const std::set<SectorGL>& sectors, //
		std::map<SectorGL, std::list<SectorGL> >& new_targets_with_equiv,
		bool minimize_target_crossings, //
		bool find_zero_sectors,//
		bool allow_crossings,//
		bool allow_general_crossings,//
		std::map<Sector, std::list<std::map<int, int> > >& cached_node_perm,
		bool verbose) {

	if (verbose)
		LOG("Finding sector shifts:");
	ProgressBar pbar(2, "sector shifts:", sectors.size());
	if (verbose) {
		pbar.start();
		LOGX("");
	}
	int num_shifts = 0;
	//map<Sector, list<map<int, int> > > symmetry_shifts;
	for (set<SectorGL>::const_iterator s = sectors.begin(); s != sectors.end(); ++s) {
		if (verbose)
			pbar.print();
		const SectorGL& sec = *s;
		set<SectorGL>::const_iterator m;
		for (m = minimal.begin(); m != minimal.end(); ++m) {
			const SectorGL& min = *m;
			const IntegralFamily* f = min.sector_.integralfamily();
			ASSERT(*f == *f->source_integralfamily()); // min not from crossed integral family
			ASSERT(min.sector_ < sec.sector_); // min lower than sec
			ASSERT(*sec.sector_.integralfamily() == *integralfamily_); // sec belong to this integral family
			Crossing crossing(f->kinematics());
			GiNaC::exmap shift;
			GiNaC::ex det(0);
			if (min.labels_.first == sec.labels_.first) {
				det = Graph::find_shift(sec.graph_, sec.labels_, min.graph_,
						min.labels_, shift, crossing,
						cached_node_perm[min.sector_]);
				Crossing inverse_crossing = crossing.inverse();
				string name = inverse_crossing.name_for_crossed_family(f);
				const Sector target = Sector(Files::instance()->integralfamily(
						name), min.sector_.id());
				const Sector& source = sec.sector_;
				if (target < source && det.is_equal(1) && (//
						allow_general_crossings //
								|| crossing.is_identity()//
								|| (allow_crossings
										&& crossing.is_equivalent_to_identity()))) {
					++num_shifts;
					LOGX("Found shift from " << source << " to " << target);
					insert_shift(source.id(), target, shift);
					// add uncrossed to new targets if it is not shifted to another family
					if (*f == *integralfamily_)
						new_targets_with_equiv[min].push_back(sec);
					else if (find_zero_sectors && is_zero(min.sector_))
						zero_sectors_.insert(source.id());
					break;
				} else if (verbose) {
					LOGX("Reject shift with |det| = " << det << " from sector "
							<< source << " to " << target);
					LOGX("  shift:    " << shift);
					LOGX("  crossing: " << crossing.name());
				}
			}
		}
		if (m == minimal.end()) { // sec not shifted to any sector in minimal
			if (verbose)
				LOGX("No shift found for sector " << sec.sector_);
			minimal.insert(minimal.end(), sec);
			if (minimize_target_crossings && sec.graph_.num_external_nodes()
					> 0 && cached_node_perm.find(sec.sector_)
					== cached_node_perm.end()) {
				cached_node_perm[sec.sector_]
						= sec.graph_.find_node_symmetry_group();
			}
			new_targets_with_equiv[sec];
		}
	}
	if (verbose) {
		pbar.end();
		LOG("  Found " << num_shifts << " new shifts to lower sectors");
	}
}

void SectorMappings::insert_zero_sectors(const std::set<int>& sector_ids) {
	zero_sectors_.insert(sector_ids.begin(), sector_ids.end());
}

int SectorMappings::find_zero_sectors(
		const std::map<Sector, std::list<Sector> >& sectors_by_rep,
		bool allow_longtest) {
	int tt = -1, col = 1, num_zeros = 0;
	map<Sector, list<Sector> >::const_iterator m;
	for (m = sectors_by_rep.begin(); m != sectors_by_rep.end(); ++m) {
		ASSERT(*(m->first.integralfamily()) == *integralfamily_);
		const Sector& s = m->first;
		int id = s.id();
		if (s.t() != tt) {
			tt = s.t();
			col = 1;
			LOGN(endl << "   t = " << tt << ":\t\b\b\b\b" << flush);
		} else if ((col++) % 60 == 0) {
			LOGN (endl << "\t\t\b\b\b\b" << flush);
		}
		if (zero_sectors_.find(id) != zero_sectors_.end())
			continue;
		set<int> sub = integralfamily_->get_subsector_equivalents(id, false);
		for (set<int>::iterator n = sub.begin(); n != sub.end();)
			if (zero_sectors_.find(*n) != zero_sectors_.end()) {
				sub.erase(n++);
			} else if (Sector(integralfamily(), *n).is_obviously_zero()) {
				zero_sectors_.insert(*n);
				sub.erase(n++);
			} else {
				break; // non-zero sub sectors for id
			}
		// calculate if zero sector (skip test if non-zero subsectors present)
		if (sub.empty() && search_corner_reduction_to_zero(id, allow_longtest)) {
			zero_sectors_.insert(id); // is zero
			// include isomorphic zero sectors
			for (list<Sector>::const_iterator z = m->second.begin(); z
					!= m->second.end(); ++z) {
				ASSERT(*(z->integralfamily()) == *integralfamily_);
				zero_sectors_.insert(z->id());
			}
			LOGN("0" << flush);
			++num_zeros;
		} else {
			LOGN("x" << flush);
		}
	}
	LOG("");
	return num_zeros;
}

void SectorMappings::find_symmetry_shifts(
		const std::set<SectorGL>& sectors,
		const std::map<Sector, std::list<std::map<int, int> > >& cached_node_perm) {
	ProgressBar pbar(2, "finding symmetry shifts", sectors.size());
	pbar.start();
	for (set<SectorGL>::const_iterator s = sectors.begin(); s != sectors.end(); ++s) {
		ASSERT(*s->sector_.integralfamily() == *integralfamily_);
		list<exmap> shifts;
		map<Sector, list<map<int, int> > >::const_iterator found;
		found = cached_node_perm.find(s->sector_);
		if (found != cached_node_perm.end())
			s->graph_.find_symmetry_shifts(shifts,
					s->sector_.integralfamily()->kinematics(), found->second);
		else
			s->graph_.find_symmetry_shifts(shifts,
					s->sector_.integralfamily()->kinematics());
		if (!shifts.empty())
			insert_symmetry_shifts(s->sector_.id(), shifts);
		pbar.print();
	}
	pbar.end();
}

bool SectorMappings::test_shift(const Sector& from, const Sector& to,
		const GiNaC::exmap& shift) {
	if (no_explicit_shift_verification_)
		return true;
	INT integral(from.integralfamily(), from.id());
	LinearCombination lhs;
	try {
		lhs = integral.apply_shift(shift, to.integralfamily(), 1, false, true);
	} catch (exception& e) {
		stringstream ss;
		ss << "shift " << shift << " doesn't map sector " << from << " to "
				<< to << "\n";
		vector<Propagator> p1 = from.get_propagators();
		vector<Propagator> p2 = to.get_propagators();
		ss << from << ": ";
		copy(p1.begin(), p1.end(), ostream_iterator<Propagator> (ss, " "));
		ss << "\n" << to << ": ";
		copy(p2.begin(), p2.end(), ostream_iterator<Propagator> (ss, " "));
		ss << "\n";
		LOG(ss.str() + e.what());
		return false;
	}
	if (lhs.size() == 1 && lhs.begin()->first.id() == to.id())
		return true;
	return false;
}

void SectorMappings::remove_shift(int from_id) {
	sector_relations_.erase(from_id);
	sector_relation_shifts_.erase(from_id);
}

void SectorMappings::insert_shift(int from_id, const Sector& to,
		const GiNaC::exmap& shift) {

	const Sector from(integralfamily_, from_id);

	// from-sectors must be minimal w.r.t. global permutation symmetries.
	int from_equiv = integralfamily_->get_equivalent(from_id);
	if (from_equiv != from_id) {
		Sector min(integralfamily_, from_equiv);
		ABORT("Reject shift from " << from << " to sector " << to << ".\n"
				<< "Sector " << from << " is equivalent to sector " << min
				<< " by global permutation symmetries.\n"
				<< "Try a shift from " << min << " to " << to << " instead.");
	}

	// target sectors must be lower sectors than "from"-sectors.
	if (!(to < from))
		ABORT("Reject shift from lower (or equal) sector " << from << " to " << to << ".\n"
				<< "Try a shift from " << to << " to " << from);

	// shifts between physical and unphysical sectors are not allowed.
	bool g1 = sectors_without_graph_.find(from_id)
			== sectors_without_graph_.end();
	const SectorMappings* tom = //
			Files::instance()->sectormappings(to.integralfamily()->name());
	bool g2 = tom->sectors_without_graph_.find(to.id())
			== tom->sectors_without_graph_.end();
	if (g1 && !g2)
		ABORT("Reject shift between physical sector " << from << " and unphysical sector " << to);
	else if (!g1 && g2)
		ABORT("Reject shift between unphysical sector " << from << " and physical sector " << to);

	if (!SectorMappings::test_shift(from, to, shift))
		ABORT("Wrong shift from sector " << from << " to sector "
				<< to << " detected: " << shift);

	pair<map<int, Sector>::const_iterator, bool> inserted1;
	pair<map<int, GiNaC::exmap>::const_iterator, bool> inserted2;
	inserted1 = sector_relations_.insert(make_pair(from_id, to));
	inserted2 = sector_relation_shifts_.insert(make_pair(from_id, shift));
	ASSERT(inserted1.second == inserted2.second);
	if (!inserted1.second)
		ABORT("Shift from " << from << " to " << to << " already inserted.");
}

void SectorMappings::insert_shift(const Sector& from, const Sector& to,
		const GiNaC::exmap& shift) {
	const string& name = from.integralfamily()->name();
	SectorMappings *m = Files::instance()->sectormappings(name);
	m->insert_shift(from.id(), to, shift);
}

void SectorMappings::insert_symmetry_shifts(int sector_id, const std::list<
		GiNaC::exmap>& shifts) {
	list<exmap>::const_iterator symm;
	Sector sec(integralfamily_, sector_id);
	for (symm = shifts.begin(); symm != shifts.end(); ++symm)
		if (!SectorMappings::test_shift(sec, sec, *symm))
			ABORT("Wrong symmetry shift for " << sec << " detected: " << *symm);
	pair<map<int, list<GiNaC::exmap> >::const_iterator, bool> inserted;
	inserted = sector_symmetries_.insert(make_pair(sector_id, shifts));
	if (!inserted.second)
		ABORT("Symmetry shifts for sector " << sec << " already inserted.");
}

void SectorMappings::insert_symmetry_shifts(const Sector& sec,
		const list<exmap>& shifts) {
	const string& name = sec.integralfamily()->name();
	SectorMappings *m = Files::instance()->sectormappings(name);
	m->insert_symmetry_shifts(sec.id(), shifts);
}

void SectorMappings::find_shift_targets(set<Sector>& sectors) const {
	int maxid = integralfamily_->max_sector_id();
	set<int> equiv = integralfamily_->get_subsector_equivalents(maxid, true);
	equiv.insert(maxid);
	// exclude sectors for which a shift exists
	for (set<int>::iterator e = equiv.begin(); e != equiv.end(); ++e)
		if (sector_relations_.find(*e) == sector_relations_.end()) // no shift found
			sectors.insert(Sector(integralfamily_, *e));
}

void SectorMappings::find_shift_targets_with_graph(
		std::set<SectorGL>& shift_targets, bool minimize_by_twists,
		bool construct_minimal_graphs) const {

	// find sectors of this family with graph and no shift
	LOG("Finding shift targets (with graph and canonical label):");
	set<Sector> sectors;
	find_shift_targets(sectors);
	for (set<Sector>::iterator s = sectors.begin(); s != sectors.end();)
		if (sectors_without_graph_.find(s->id())
				!= sectors_without_graph_.end())
			sectors.erase(s++);
		else
			++s;

	// construct the graphs
	map<Sector, Graph> graph_by_sec;
	get_graphs(graph_by_sec, sectors, minimize_by_twists,
			construct_minimal_graphs, true);

	// calculate the canonical label
	for (map<Sector, Graph>::const_iterator s = graph_by_sec.begin(); s
			!= graph_by_sec.end(); ++s) {
		shift_targets.insert(SectorGL(s->first, s->second,
				s->second.find_canonical_graph_label()));
	}
}

/// \todo: idea for improvement:
// show separate t graph: 4->3->2->1->0 vertically
// and show sectors at appropriate vert. position
// similarly to timeline example on http://www.graphviz.org/pdf/dotguide.pdf
// page 16/17
void SectorMappings::print_sectortree_dot(const std::set<Sector>& secs,
		const std::string& filename) {
	string filename_tmp = filename + ".tmp";
	ofstream file(filename_tmp.c_str());

	// header
	file << "digraph SectorRelations {\n";
	file << "size = \"8.3,11.7\"\n";
	file << "ratio = \"1.5\"\n";
	file << "center = \"true\"\n";

	set<Sector> allsecs = secs;
	set<Sector>::const_iterator s;
	for (s = secs.begin(); s != secs.end(); ++s) {
		INT ie(SectorMappings::get_equivalent_or_unique_zero(INT(*s)));
		Sector equiv = ie.get_sector();
		if (equiv == *s)
			continue;
		allsecs.insert(equiv);
		file//
				<< s->integralfamily()->name()//
				<< s->id()//
				<< " -> "//
				<< equiv.integralfamily()->name()//
				<< equiv.id()//
				<< " [color=green];\n";
	}
	for (s = secs.begin(); s != secs.end(); ++s) {
		set<Sector> sub = SectorSelection::find_prerequisite_sectors(*s, true);
		allsecs.insert(sub.begin(), sub.end());
	}

	// sectors are vertices, relations are directed edges
	map<Sector, set<Sector> > relations;
	for (s = allsecs.begin(); s != allsecs.end(); ++s)
		relations[*s] = SectorSelection::find_prerequisite_sectors(*s, false);

	map<Sector, set<Sector> >::const_iterator r;
	for (r = relations.begin(); r != relations.end(); ++r) {
		for (s = r->second.begin(); s != r->second.end(); ++s)
			file//
					<< r->first.integralfamily()->name() << r->first.id()//
					<< " -> "//
					<< s->integralfamily()->name()//
					<< s->id()//
					<< " [color=blue];\n";
	}
	for (s = allsecs.begin(); s != allsecs.end(); ++s)
		file << s->integralfamily()->name()//
				<< s->id()//
				<< " [label=\"" << *s
		/*<< " " << s->t()*/<< "\"];\n";

	// trailer
	file << "}\n";

	file.close();
	rename(filename_tmp, filename);
}

void read_sector_relations(const YAML::Node& node, const IntegralFamily* fam,
		map<int, Sector>& relations, map<int, exmap>& shifts) {
	using namespace YAML;
	for (Iterator r = node.begin(); r != node.end(); ++r) {
		int from_equiv_id;
		r.first() >> from_equiv_id;
		Sector from_equiv_sec(fam, from_equiv_id);
		Sector to_equiv_sec(r.second()[0]);

		bool new_notation = true;
		if (r.second().size() == 3) {
			new_notation = false;
			int from_id;
			r.second()[1] >> from_id;
			Sector from_sec(fam, from_id);
			VERIFY(from_sec == from_equiv_sec);
		} else {
			VERIFY(r.second().size() == 2);
		}

		const Node& shift_node = new_notation ? r.second()[1] : r.second()[2];
		exmap shift;
		lst from_momenta = fam->get_all_momenta();
		lst to_momenta = to_equiv_sec.integralfamily()->get_all_momenta();
		for (Iterator s = shift_node.begin(); s != shift_node.end(); ++s) {
			string lhs, rhs;
			(*s)[0] >> lhs;
			(*s)[1] >> rhs;
			shift[ex(lhs, from_momenta)] = ex(rhs, to_momenta);
		}
		if (!SectorMappings::test_shift(from_equiv_sec, to_equiv_sec, shift))
			ABORT("Wrong shift from sector " << from_equiv_sec << " to sector "
					<< to_equiv_sec << " detected: " << shift);
		relations.insert(make_pair(from_equiv_id, to_equiv_sec));
		shifts.insert(make_pair(from_equiv_id, shift));
	}
}

void operator >>(const YAML::Node& node, SectorMappings& m) {
	using namespace YAML;
	using namespace GiNaC;
	string name;
	node["name"] >> name;
	string required = m.integralfamily_->name();
	if (name != required)
		throw runtime_error("Name " + name + " does not match " + required);

	const Node& z_node = node["zero_sectors"];
	for (Iterator n = z_node.begin(); n != z_node.end(); ++n)
		for (Iterator s = n.second().begin(); s != n.second().end(); ++s)
			m.zero_sectors_.insert(*s);

	if (node.FindValue("sectors_without_graph")) {
		const Node& ng_node = node["sectors_without_graph"];
		for (Iterator n = ng_node.begin(); n != ng_node.end(); ++n)
			for (Iterator s = n.second().begin(); s != n.second().end(); ++s)
				m.sectors_without_graph_.insert(*s);
	}

	const Node& rel_node = node["sector_relations"];
	read_sector_relations(rel_node, m.integralfamily_, m.sector_relations_,
			m.sector_relation_shifts_);

	if (node.FindValue("crossed_sector_relations")) {
		const Node& xrel_node = node["crossed_sector_relations"];
		for (Iterator r = xrel_node.begin(); r != xrel_node.end(); ++r) {
			string xname;
			r.first() >> xname;
			string xfamname = m.integralfamily()->name() + xname;
			SectorMappings* xm = Files::instance()->sectormappings(xfamname);
			read_sector_relations(r.second(), xm->integralfamily_,
					xm->sector_relations_, xm->sector_relation_shifts_);
		}
	}

	if (node.FindValue("sector_symmetries")) {
		const Node& sym_node = node["sector_symmetries"];
		const lst& symbols = m.integralfamily_->get_all_momenta();
		for (Iterator r = sym_node.begin(); r != sym_node.end(); ++r) {
			int id;
			r.first() >> id;
			Sector sector(m.integralfamily_, id);
			list<exmap> id_shifts;
			for (Iterator s = r.second().begin(); s != r.second().end(); ++s) {
				exmap shift;
				read(*s, shift, symbols);
				YAML::Emitter ye;
				ye << Flow << shift;
				if (!SectorMappings::test_shift(sector, sector, shift))
					ABORT("Wrong symmetry shift for topology " << id << " detected: " << ye.c_str());
				id_shifts.push_back(shift);
			}
			m.sector_symmetries_.insert(make_pair(id, id_shifts));
		}
	}
}

void write_sector_relations(YAML::Emitter& os,
		const map<int, Sector>& relations, const map<int, exmap>& shifts) {
	using namespace YAML;
	os << BeginMap;
	map<int, GiNaC::exmap>::const_iterator r;
	for (r = shifts.begin(); r != shifts.end(); ++r) {
		int from_equiv = r->first;
		const exmap& shift = r->second;
		map<int, Sector>::const_iterator equiv = relations.find(from_equiv);
		VERIFY(equiv != relations.end());
		const Sector& to_equiv = equiv->second;
		os << Key << from_equiv << Value << Flow << BeginSeq;
		os << to_equiv;
		os << BeginSeq;
		for (exmap::const_iterator sr = shift.begin(); sr != shift.end(); ++sr)
			os << BeginSeq << sr->first << sr->second << EndSeq;
		os << EndSeq;
		os << EndSeq;
	}
	os << EndMap;
}

YAML::Emitter & operator<<(YAML::Emitter& os, const SectorMappings& m) {
	using namespace YAML;
	using namespace GiNaC;
	const IntegralFamily* ic = m.integralfamily();
	os << BeginMap;

	os << Key << "name" << Value << ic->name();

	os << Key << "zero_sectors" << Value << BeginMap;
	map<int, set<int> > z_by_t; // collect by common t
	set<int>::const_iterator z;
	for (z = m.zero_sectors_.begin(); z != m.zero_sectors_.end(); ++z)
		z_by_t[ic->get_t(*z)].insert(*z);
	map<int, set<int> >::iterator i;
	for (i = z_by_t.begin(); i != z_by_t.end(); ++i) {
		stringstream ss;
		ss << "t=" << i->first;
		os << Key << ss.str() << Value << Flow << BeginSeq;
		for (set<int>::iterator s = i->second.begin(); s != i->second.end(); ++s)
			os << *s;
		os << EndSeq;
	}
	os << EndMap;

	os << Key << "sectors_without_graph" << Value << BeginMap;
	map<int, set<int> > ng_by_t; // collect by common t
	set<int>::const_iterator seci;
	for (seci = m.sectors_without_graph_.begin(); seci
			!= m.sectors_without_graph_.end(); ++seci)
		ng_by_t[ic->get_t(*seci)].insert(*seci);
	//map<int, set<int> >::iterator i;
	for (i = ng_by_t.begin(); i != ng_by_t.end(); ++i) {
		stringstream ss;
		ss << "t=" << i->first;
		os << Key << ss.str() << Value << Flow << BeginSeq;
		for (set<int>::iterator s = i->second.begin(); s != i->second.end(); ++s)
			os << *s;
		os << EndSeq;
	}
	os << EndMap;

	os << Key << "sector_relations" << Value;
	write_sector_relations(os, m.sector_relations_, m.sector_relation_shifts_);

	os << Key << "crossed_sector_relations" << Value << BeginMap;
	list<Crossing> xs = m.integralfamily()->kinematics()->ordered_crossings();
	for (list<Crossing>::const_iterator x = xs.begin(); x != xs.end(); ++x) {
		string fn = x->name_for_crossed_family(m.integralfamily());
		SectorMappings* xm = Files::instance()->sectormappings(fn);
		if (!xm->sector_relations_.empty()) {
			os << Key << x->name() << Value;
			write_sector_relations(os, xm->sector_relations_,
					xm->sector_relation_shifts_);
		}
	}
	os << EndMap;

	os << Key << "sector_symmetries" << Value << BeginMap;
	const map<int, list<GiNaC::exmap> >& syms = m.sector_symmetries_;
	map<int, list<GiNaC::exmap> >::const_iterator it;
	list<GiNaC::exmap>::const_iterator it2;
	for (it = syms.begin(); it != syms.end(); ++it) {
		os << Key << it->first << Value;
		os << BeginSeq;
		for (it2 = it->second.begin(); it2 != it->second.end(); ++it2) {
			os << Flow << *it2;
		}
		os << EndSeq;
	}
	os << EndMap;

	os << EndMap;
	return os;
}

void SectorMappings::verify_is_equivalent(const SectorMappings& other) const {
	try {
		if (integralfamily()->name() != other.integralfamily()->name())
			throw runtime_error("integral family names differ");
		if (zero_sectors() != other.zero_sectors())
			throw runtime_error("zero sectors differ");
		if (sectors_without_graph() != other.sectors_without_graph())
			throw runtime_error("sectors without graph differ");
		const map<int, Sector>& secrel1 = sector_relations();
		const map<int, Sector>& secrel2 = other.sector_relations();
		if (secrel1 != secrel2)
			throw runtime_error("sector relations differ");

		const map<int, list<GiNaC::exmap> >& sym1 = sector_symmetries();
		const map<int, list<GiNaC::exmap> >& sym2 = other.sector_symmetries();
		if (sym1.size() != sym2.size())
			throw runtime_error("not the same number of sector symmetries");
		map<int, list<GiNaC::exmap> >::const_iterator it1 = sym1.begin(), it2 =
				sym2.begin();
		int lauf = 0;
		for (; it1 != sym1.end(); ++it1, ++it2) {
			++lauf;
			if (it1->first != it2->first || it1->second.size()
					!= it2->second.size())
				throw runtime_error("sector relation mismatch at position "
						+ to_string(lauf));
		}
	} catch (exception& e) {
		throw runtime_error(integralfamily_->name() + ": " + e.what());
	}
}

}
// namespace Reduze
